instance_profile = "arn:aws:iam::018967853140:instance-profile/data-access-profile"

s3_log_uri = "s3n://aws-logs-018967853140-us-east-1/elasticmapreduce/"

master_security_group = "sg-0f9b7e8c1d91cb290"
slave_security_group = "sg-0d69b38ec2d7da5db"
service_role = "service-role/AmazonEMR-ServiceRole-20240117T104113"
job_role = "arn:aws:iam::018967853140:instance-profile/AmazonEMR-InstanceProfile-20240117T104057"
aws_region = "us-east-1"
rackspace_user = "018967853140"
